import { useRef, useState, useEffect, createContext, createElement, useCallback, useDebugValue, useContext } from 'react';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

const hasInitialValue = atom => 'init' in atom;

const concatMap = (m1, m2) => {
  // Map([...m1, ...m2]) alternative
  const newMap = new Map(m1);
  m2.forEach((value, key) => {
    newMap.set(key, value);
  });
  return newMap;
};

const createState = (initialValues, newAtomReceiver) => {
  let currWip = new Map();
  const queue = [];

  const updateState = updater => {
    queue.push(updater);

    if (queue.length > 1) {
      return;
    }

    let nextWip = currWip;

    while (queue.length) {
      nextWip = queue[0](nextWip);
      queue.shift();
    }

    if (nextWip !== currWip) {
      currWip = nextWip;

      if (currWip.size) {
        const atomsToNotify = new Set(currWip.keys());
        mountDependencies(state, currWip);
        commitState(state, currWip);
        ++state.v;
        atomsToNotify.forEach(atom => {
          const mounted = state.m.get(atom);
          mounted == null ? void 0 : mounted.l.forEach(listener => listener());
        });
      }
    }
  };

  const state = {
    n: newAtomReceiver,
    v: 0,
    a: new WeakMap(),
    m: new WeakMap(),
    u: updateState
  };

  if (initialValues) {
    for (const [atom, value] of initialValues) {
      const atomState = {
        v: value,
        r: 0,
        d: new Map()
      };

      if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
        Object.freeze(atomState);
      }

      state.a.set(atom, atomState);
    }
  }

  return state;
};

const getAtomState = (state, wip, atom) => wip.get(atom) || state.a.get(atom);

const wipAtomState = (state, wip, atom, dependencies, promise) => {
  const atomState = getAtomState(state, wip, atom);

  if (promise && promise !== (atomState == null ? void 0 : atomState.p)) {
    // newer async read is running, not updating
    return null;
  }

  const nextAtomState = _extends({
    r: 0
  }, atomState, {
    d: dependencies ? new Map(Array.from(dependencies).map(a => {
      var _getAtomState$r, _getAtomState;

      return [a, (_getAtomState$r = (_getAtomState = getAtomState(state, wip, a)) == null ? void 0 : _getAtomState.r) != null ? _getAtomState$r : 0];
    })) : atomState ? atomState.d : new Map()
  });

  if (!atomState && hasInitialValue(atom)) {
    nextAtomState.v = atom.init;
  }

  return nextAtomState;
};

const setAtomValue = (state, wip, atom, value, dependencies, promise) => {
  const atomState = wipAtomState(state, wip, atom, dependencies, promise);

  if (!atomState) {
    return wip;
  }

  delete atomState.e;
  delete atomState.p;

  if (!('v' in atomState) || !Object.is(atomState.v, value)) {
    atomState.v = value;
    ++atomState.r;
  }

  return new Map(wip).set(atom, atomState);
};

const setAtomReadError = (state, wip, atom, error, dependencies, promise) => {
  const atomState = wipAtomState(state, wip, atom, dependencies, promise);

  if (!atomState) {
    return wip;
  }

  delete atomState.p;
  atomState.e = error;
  return new Map(wip).set(atom, atomState);
};

const setAtomReadPromise = (state, wip, atom, promise, dependencies) => {
  const atomState = wipAtomState(state, wip, atom, dependencies);
  atomState.p = promise;
  return new Map(wip).set(atom, atomState);
};

const setAtomWritePromise = (state, wip, atom, promise) => {
  const atomState = wipAtomState(state, wip, atom);

  if (promise) {
    atomState.w = promise;
  } else {
    delete atomState.w;
  }

  return new Map(wip).set(atom, atomState);
};

const scheduleReadAtomState = (state, atom, promise) => {
  promise.then(() => {
    state.u(wip => readAtomState(state, wip, atom, true)[1]);
  });
};

const readAtomState = (state, wip, atom, force) => {
  if (!force) {
    const atomState = getAtomState(state, wip, atom);

    if (atomState && Array.from(atomState.d.entries()).every(([a, r]) => {
      const aState = getAtomState(state, wip, a);
      return aState && !aState.e && !aState.p && aState.r === r;
    })) {
      return [atomState, wip];
    }
  }

  let error;
  let promise;
  let value;
  const dependencies = new Set();

  try {
    const promiseOrValue = atom.read(a => {
      dependencies.add(a);

      if (a !== atom) {
        let aState;
        [aState, wip] = readAtomState(state, wip, a);

        if (aState.e) {
          throw aState.e; // read error
        }

        if (aState.p) {
          throw aState.p; // read promise
        }

        return aState.v; // value
      } // a === atom


      const aState = getAtomState(state, wip, a);

      if (aState) {
        if (aState.p) {
          throw aState.p; // read promise
        }

        return aState.v; // value
      }

      if (hasInitialValue(a)) {
        return a.init;
      }

      throw new Error('no atom init');
    });

    if (promiseOrValue instanceof Promise) {
      promise = promiseOrValue.then(value => {
        state.u(prev => setAtomValue(state, concatMap(prev, wip), atom, value, dependencies, promise));
      }).catch(e => {
        if (e instanceof Promise) {
          scheduleReadAtomState(state, atom, e);
          return e;
        }

        state.u(prev => setAtomReadError(state, concatMap(prev, wip), atom, e instanceof Error ? e : new Error(e), dependencies, promise));
      });
    } else {
      value = promiseOrValue;
    }
  } catch (errorOrPromise) {
    if (errorOrPromise instanceof Promise) {
      scheduleReadAtomState(state, atom, errorOrPromise);
      promise = errorOrPromise;
    } else if (errorOrPromise instanceof Error) {
      error = errorOrPromise;
    } else {
      error = new Error(errorOrPromise);
    }
  }

  if (error) {
    wip = setAtomReadError(state, wip, atom, error, dependencies);
  } else if (promise) {
    wip = setAtomReadPromise(state, wip, atom, promise, dependencies);
  } else {
    wip = setAtomValue(state, wip, atom, value, dependencies);
  }

  const ret = [getAtomState(state, wip, atom), wip];
  wip = new Map(); // for async

  return ret;
};

const readAtom = (state, readingAtom) => {
  const [atomState, wip] = readAtomState(state, new Map(), readingAtom); // schedule commit

  if (wip.size) {
    state.u(prev => {
      // XXX this is very tricky, any idea to improve?
      mountDependencies(state, wip);
      commitState(state, wip);
      return prev;
    });
  }

  return atomState;
};

const addAtom = (state, addingAtom) => {
  let mounted = state.m.get(addingAtom);

  if (!mounted) {
    mounted = mountAtom(state, new Map(), addingAtom);
  }

  return mounted;
}; // XXX doesn't work with mutally dependent atoms


const canUnmountAtom = (atom, mounted) => !mounted.l.size && (!mounted.d.size || mounted.d.size === 1 && mounted.d.has(atom));

const delAtom = (state, deletingAtom) => {
  const mounted = state.m.get(deletingAtom);

  if (mounted && canUnmountAtom(deletingAtom, mounted)) {
    unmountAtom(state, new Map(), deletingAtom);
  }
};

const updateDependentsState = (state, wip, atom, prevAtomState) => {
  var _getAtomState2;

  if (prevAtomState && !prevAtomState.e && !prevAtomState.p && prevAtomState.r === ((_getAtomState2 = getAtomState(state, wip, atom)) == null ? void 0 : _getAtomState2.r)) {
    return wip; // bail out
  }

  const mounted = state.m.get(atom);
  mounted == null ? void 0 : mounted.d.forEach(dependent => {
    var _nextDependentState$p;

    if (dependent === atom) {
      return;
    }

    const dependentState = getAtomState(state, wip, dependent);
    let nextDependentState;
    [nextDependentState, wip] = readAtomState(state, wip, dependent, true);
    (_nextDependentState$p = nextDependentState.p) == null ? void 0 : _nextDependentState$p.then(() => {
      state.u(prev => updateDependentsState(state, prev, dependent));
    });
    wip = updateDependentsState(state, wip, dependent, dependentState);
  });
  return wip;
};

const writeAtomState = (state, wip, atom, update, pendingPromises) => {
  const atomState = getAtomState(state, wip, atom);

  if (atomState && atomState.w) {
    const promise = atomState.w.then(() => {
      state.u(prev => writeAtomState(state, prev, atom, update));
    });

    if (pendingPromises) {
      pendingPromises.push(promise);
    }

    return wip;
  }

  let isSync = true;

  try {
    const promiseOrVoid = atom.write(a => {
      const [aState] = readAtomState(state, wip, a);

      if (aState.e) {
        throw aState.e; // read error
      }

      if (aState.p) {
        if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
          console.warn('Reading pending atom state in write operation. We throw a promise for now.', a);
        }

        throw aState.p; // read promise
      }

      if ('v' in aState) {
        return aState.v;
      }

      if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
        console.warn('[Bug] no value found while reading atom in write operation. This probably a bug.', a);
      }

      throw new Error('no value found');
    }, (a, v) => {
      if (a === atom) {
        const aState = getAtomState(state, wip, a);

        if (isSync) {
          wip = updateDependentsState(state, setAtomValue(state, wip, a, v), a, aState);
        } else {
          state.u(prev => updateDependentsState(state, setAtomValue(state, prev, a, v), a, aState));
        }
      } else {
        if (isSync) {
          wip = writeAtomState(state, wip, a, v);
        } else {
          state.u(prev => writeAtomState(state, prev, a, v));
        }
      }
    }, update);

    if (promiseOrVoid instanceof Promise) {
      if (pendingPromises) {
        pendingPromises.push(promiseOrVoid);
      }

      wip = setAtomWritePromise(state, wip, atom, promiseOrVoid.then(() => {
        state.u(prev => setAtomWritePromise(state, prev, atom));
      }));
    }
  } catch (e) {
    if (pendingPromises && pendingPromises.length) {
      pendingPromises.push(new Promise((_resolve, reject) => {
        reject(e);
      }));
    } else {
      throw e;
    }
  }

  isSync = false;
  return wip;
};

const writeAtom = (state, writingAtom, update) => {
  const pendingPromises = [];
  state.u(prev => writeAtomState(state, prev, writingAtom, update, pendingPromises));

  if (pendingPromises.length) {
    return new Promise((resolve, reject) => {
      const loop = () => {
        const len = pendingPromises.length;

        if (len === 0) {
          resolve();
        } else {
          Promise.all(pendingPromises).then(() => {
            pendingPromises.splice(0, len);
            loop();
          }).catch(reject);
        }
      };

      loop();
    });
  }
};

const isActuallyWritableAtom = atom => !!atom.write;

const mountAtom = (state, wip, atom, initialDependent) => {
  // mount dependencies beforehand
  const atomState = getAtomState(state, wip, atom);

  if (atomState) {
    atomState.d.forEach((_, a) => {
      if (a !== atom) {
        // check if not mounted
        if (!state.m.has(a)) {
          mountAtom(state, wip, a, atom);
        }
      }
    });
  } else if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
    console.warn('[Bug] could not find atom state to mount', atom);
  } // mount self


  const mounted = {
    d: new Set(initialDependent && [initialDependent]),
    l: new Set(),
    u: undefined
  };
  state.m.set(atom, mounted);
  state.u(prev => {
    if (isActuallyWritableAtom(atom) && atom.onMount) {
      const setAtom = update => writeAtom(state, atom, update);

      mounted.u = atom.onMount(setAtom);
    }

    return prev;
  });
  return mounted;
};

const unmountAtom = (state, wip, atom) => {
  var _state$m$get;

  // unmount self
  const onUnmount = (_state$m$get = state.m.get(atom)) == null ? void 0 : _state$m$get.u;

  if (onUnmount) {
    onUnmount();
  }

  state.m.delete(atom); // unmount dependencies afterward

  const atomState = getAtomState(state, wip, atom);

  if (atomState) {
    if (atomState.p && typeof process === 'object' && process.env.NODE_ENV !== 'production') {
      console.warn('[Bug] deleting atomState with read promise', atom);
    }

    atomState.d.forEach((_, a) => {
      if (a !== atom) {
        const mounted = state.m.get(a);

        if (mounted) {
          mounted.d.delete(atom);

          if (canUnmountAtom(a, mounted)) {
            unmountAtom(state, wip, a);
          }
        }
      }
    });
  } else if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
    console.warn('[Bug] could not find atom state to unmount', atom);
  }
};

const mountDependencies = (state, wip) => {
  wip.forEach((atomState, atom) => {
    var _state$a$get;

    const prevDependencies = (_state$a$get = state.a.get(atom)) == null ? void 0 : _state$a$get.d;

    if (prevDependencies === atomState.d) {
      return;
    }

    const dependencies = new Set(atomState.d.keys());

    if (prevDependencies) {
      prevDependencies.forEach((_, a) => {
        const mounted = state.m.get(a);

        if (dependencies.has(a)) {
          // not changed
          dependencies.delete(a);
        } else if (mounted) {
          mounted.d.delete(atom);

          if (canUnmountAtom(a, mounted)) {
            unmountAtom(state, wip, a);
          }
        } else if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
          console.warn('[Bug] a dependency is not mounted', a);
        }
      });
    }

    dependencies.forEach(a => {
      const mounted = state.m.get(a);

      if (mounted) {
        const dependents = mounted.d;
        dependents.add(atom);
      } else {
        mountAtom(state, wip, a, atom);
      }
    });
  });
};

const commitState = (state, wip) => {
  // copy wip to AtomStateMap
  wip.forEach((atomState, atom) => {
    if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
      Object.freeze(atomState);
    }

    const isNewAtom = state.n && !state.a.has(atom);
    state.a.set(atom, atomState);

    if (isNewAtom) {
      state.n(atom);
    }
  }); // empty wip

  wip.clear();
};

const subscribeAtom = (state, atom, callback) => {
  const mounted = addAtom(state, atom);
  const listeners = mounted.l;
  listeners.add(callback);
  return () => {
    listeners.delete(callback);
    delAtom(state, atom);
  };
};

/*
export {
  unstable_createMutableSource as createMutableSource,
  unstable_useMutableSource as useMutableSource,
} from 'react'
*/
const TARGET = Symbol();
const GET_VERSION = Symbol();
const createMutableSource = (target, getVersion) => ({
  [TARGET]: target,
  [GET_VERSION]: getVersion
});
const useMutableSource = (source, getSnapshot, subscribe) => {
  const lastVersion = useRef(0);
  const currentVersion = source[GET_VERSION](source[TARGET]);
  const [state, setState] = useState(() => [
  /* [0] */
  source,
  /* [1] */
  getSnapshot,
  /* [2] */
  subscribe,
  /* [3] */
  currentVersion,
  /* [4] */
  getSnapshot(source[TARGET])]);
  let currentSnapshot = state[4];

  if (state[0] !== source || state[1] !== getSnapshot || state[2] !== subscribe || currentVersion !== state[3] && currentVersion !== lastVersion.current) {
    currentSnapshot = getSnapshot(source[TARGET]);
    setState([
    /* [0] */
    source,
    /* [1] */
    getSnapshot,
    /* [2] */
    subscribe,
    /* [3] */
    currentVersion,
    /* [4] */
    currentSnapshot]);
  }

  useEffect(() => {
    let didUnsubscribe = false;

    const checkForUpdates = () => {
      if (didUnsubscribe) {
        return;
      }

      try {
        const nextSnapshot = getSnapshot(source[TARGET]);
        const nextVersion = source[GET_VERSION](source[TARGET]);
        lastVersion.current = nextVersion;
        setState(prev => {
          if (prev[0] !== source || prev[1] !== getSnapshot || prev[2] !== subscribe) {
            return prev;
          }

          if (prev[4] === nextSnapshot) {
            return prev;
          }

          return [
          /* [0] */
          prev[0],
          /* [1] */
          prev[1],
          /* [2] */
          prev[2],
          /* [3] */
          nextVersion,
          /* [4] */
          nextSnapshot];
        });
      } catch (e) {
        // schedule update
        setState(prev => [...prev]);
      }
    };

    const unsubscribe = subscribe(source[TARGET], checkForUpdates);
    checkForUpdates();
    return () => {
      didUnsubscribe = true;
      unsubscribe();
    };
  }, [source, getSnapshot, subscribe]);
  return currentSnapshot;
};

const createStore = (initialValues, newAtomReceiver) => {
  const state = createState(initialValues, newAtomReceiver);
  const mutableSource = createMutableSource(state, () => state.v);

  const updateAtom = (atom, update) => writeAtom(state, atom, update);

  return [mutableSource, updateAtom];
}; // export only for tests/error.test.tsx, FIXME why do we need this?

const StoreContextMap = new Map();
const getStoreContext = scope => {
  if (!StoreContextMap.has(scope)) {
    StoreContextMap.set(scope, createContext(createStore()));
  }

  return StoreContextMap.get(scope);
};

const Provider = ({
  initialValues,
  scope,
  children
}) => {
  const storeRef = useRef(null);

  if (typeof process === 'object' && process.env.NODE_ENV !== 'production' && process.env.NODE_ENV !== 'test') {
    /* eslint-disable react-hooks/rules-of-hooks */
    const atomsRef = useRef([]);

    if (storeRef.current === null) {
      // lazy initialization
      storeRef.current = createStore(initialValues, newAtom => {
        atomsRef.current.push(newAtom);
      });
    }

    useDebugState(storeRef.current, atomsRef.current);
    /* eslint-enable react-hooks/rules-of-hooks */
  } else {
    if (storeRef.current === null) {
      // lazy initialization
      storeRef.current = createStore(initialValues);
    }
  }

  const StoreContext = getStoreContext(scope);
  return createElement(StoreContext.Provider, {
    value: storeRef.current
  }, children);
};

const atomToPrintable = atom => atom.debugLabel || atom.toString();

const stateToPrintable = ([state, atoms]) => Object.fromEntries(atoms.flatMap(atom => {
  const mounted = state.m.get(atom);

  if (!mounted) {
    return [];
  }

  const dependents = mounted.d;
  const atomState = state.a.get(atom) || {};
  return [[atomToPrintable(atom), {
    value: atomState.e || atomState.p || atomState.w || atomState.v,
    dependents: Array.from(dependents).map(atomToPrintable)
  }]];
}));

const getState = state => _extends({}, state); // shallow copy
// We keep a reference to the atoms in Provider's atomsRef in dev mode,
// so atoms aren't garbage collected by the WeakMap of mounted atoms


const useDebugState = (store, atoms) => {
  const subscribe = useCallback((state, callback) => {
    // FIXME we don't need to resubscribe, just need to subscribe for new one
    const unsubs = atoms.map(atom => subscribeAtom(state, atom, callback));
    return () => {
      unsubs.forEach(unsub => unsub());
    };
  }, [atoms]);
  const state = useMutableSource(store[0], getState, subscribe);
  useDebugValue([state, atoms], stateToPrintable);
};

let keyCount = 0; // global key count for all atoms

function atom(read, write) {
  const key = `atom${++keyCount}`;
  const config = {
    toString: () => key
  };

  if (typeof read === 'function') {
    config.read = read;
  } else {
    config.init = read;

    config.read = get => get(config);

    config.write = (get, set, update) => {
      set(config, typeof update === 'function' ? update(get(config)) : update);
    };
  }

  if (write) {
    config.write = write;
  }

  return config;
}

const isWritable = atom => !!atom.write;

function useAtom(atom) {
  const getAtomValue = useCallback(state => {
    const atomState = readAtom(state, atom);

    if (atomState.e) {
      throw atomState.e; // read error
    }

    if (atomState.p) {
      throw atomState.p; // read promise
    }

    if (atomState.w) {
      throw atomState.w; // write promise
    }

    if ('v' in atomState) {
      return atomState.v;
    }

    throw new Error('no atom value');
  }, [atom]);
  const subscribe = useCallback((state, callback) => subscribeAtom(state, atom, callback), [atom]);
  const StoreContext = getStoreContext(atom.scope);
  const [mutableSource, updateAtom] = useContext(StoreContext);
  const value = useMutableSource(mutableSource, getAtomValue, subscribe);
  const setAtom = useCallback(update => {
    if (isWritable(atom)) {
      return updateAtom(atom, update);
    } else {
      throw new Error('not writable atom');
    }
  }, [updateAtom, atom]);
  useDebugValue(value);
  return [value, setAtom];
}

export { Provider, getStoreContext as SECRET_INTERNAL_getStoreContext, atom, useAtom };
